#! /usr/bin/env python3

# dbnomics-data-model -- Define, validate and transform DBnomics data.
# By: Christophe Benz <christophe.benz@cepremap.org>
#
# Copyright (C) 2017-2018 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-data-model
#
# dbnomics-data-model is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-data-model is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Validate a DBnomics storage directory containing data from a provider."""

import argparse
import io
import logging
import sys
from pathlib import Path

import jsonschema.exceptions
from toolz import get, take

import ujson as json
from dbnomics_data_model import storages, validators
from dbnomics_data_model.exceptions import StorageError
from dbnomics_data_model.observations import NOT_AVAILABLE, detect_period_format_strict, value_to_float

args = None  # Will be defined by main().
log = logging.getLogger(__name__)


def main():
    global args
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('storage_dir', type=Path, help='path of directory '
                        'containing DBnomics series from a provider (in JSON and TSV formats)')
    parser.add_argument('--format', choices=["json", "text"], default="text", help="format of the error report")
    parser.add_argument('--ignore-errors',
                        help='codes of errors to ignore, comma-separated if multiple'
                        ' (example: --ignore-errors invalid-provider-directory-name)')
    parser.add_argument('--log', default='WARNING', help='level of logging messages')
    parser.add_argument('--all-series', action='store_true', help='validate all series')
    parser.add_argument('--max-series', type=int, default=100,
                        help='Max number of series to validate per dataset')
    parser.add_argument('--all-observations', action='store_true', help='validate all observations')
    parser.add_argument('--max-observations', type=int, default=100,
                        help='Max number of observations to validate per series')
    args = parser.parse_args()

    numeric_level = getattr(logging, args.log.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: {}'.format(args.log))
    logging.basicConfig(
        format="%(levelname)s:%(name)s:%(message)s",
        level=numeric_level,
        stream=sys.stderr,
    )

    args.ignore_errors = args.ignore_errors.strip().split(",") \
        if args.ignore_errors is not None \
        else []
    if args.all_series:
        args.max_series = None
    if args.all_observations:
        args.max_observations = None

    log.debug("Loading storage...")
    try:
        storage = storages.init_storage(args.storage_dir)
    except StorageError as exc:
        print(format_error({
            **error_from_storage_error(exc),
            "location": ".",
        }))
        return -1

    log.debug("Validating provider...")
    _, provider_errors = validate_provider(storage)
    for error in provider_errors:
        print(format_error(error))

    log.debug("Validating category tree...")
    category_tree_errors = validate_category_tree(storage)
    for error in category_tree_errors:
        print(format_error(error))

    log.debug("Validating datasets...")
    nb_datasets = storage.get_nb_datasets()
    for dataset_index, dataset_dir in enumerate(storage.iter_datasets_dirs(sort_by_dir_name=True), start=1):
        dataset_code = dataset_dir.dataset_code

        log.debug("Validating dataset %s (%d/%d) (except its series)...", dataset_code, dataset_index, nb_datasets)
        _, dataset_series, dataset_errors = validate_dataset(dataset_dir)
        for error in dataset_errors:
            print(format_error(error))

        log.debug("Validating series of dataset %r...", dataset_code)
        _, series_errors = validate_series(dataset_dir, dataset_series)
        for error in series_errors:
            print(format_error(error))

    return 0


def error_from_storage_error(exc: StorageError):
    return {
        "message": exc.base_message,
        "provider_code": exc.provider_code,
        "dataset_code": exc.dataset_code,
        "series_code": exc.series_code,
    }


def format_error(error):
    global args
    if args.format == "json":
        return json.dumps(error, sort_keys=True)

    sio = io.StringIO()

    sio.write("- ")

    provider_code = error.get("provider_code")
    dataset_code = error.get("dataset_code")
    series_code = error.get("series_code")
    if provider_code is not None and dataset_code is not None and series_code is not None:
        sio.write('Series "{}/{}/{}" '.format(provider_code, dataset_code, series_code))
    elif provider_code is not None and dataset_code is not None:
        sio.write('Dataset "{}/{}" '.format(provider_code, dataset_code))
    elif provider_code is not None:
        sio.write('Provider "{}" '.format(provider_code))

    location = error.get("location")
    if location is not None:
        sio.write("at location {} ".format(location))

    line_number = error.get("line_number")
    if line_number is not None:
        sio.write("(line {}) ".format(line_number))

    sio.write("\n\n")

    sio.write("    Error message: " + error["message"] + "\n")

    sio.write("\n")

    cause = error.get("cause")
    if cause is not None:
        sio.write("    Cause:\n")
        for cause_error in cause:
            sio.write("      - at path: {}\n".format(cause_error["path"]))
            sio.write("        message: {}\n".format(cause_error["message"]))
            sio.write("        value: {}\n".format(cause_error["value"]))

        sio.write("\n")

    error_code = error.get("error_code")
    if error_code is not None:
        sio.write("    Error code: {}".format(error_code))

    sio.write("\n\n")

    return sio.getvalue()


def build_jsonschema_error(errors, base_path=[]):
    # From https://python-jsonschema.readthedocs.io/en/latest/errors/
    return [
        {
            "path": base_path + list(error.path),
            "message": error.message,
            "value": error.instance,
        }
        for error in sorted(errors, key=jsonschema.exceptions.relevance)
    ]


def validate_category_tree(storage):
    provider_code = storage.provider_code
    errors = []

    log.debug("Loading category_tree.json...")
    category_tree_json = storage.load_category_tree_json()
    if category_tree_json is not None:
        error_code = "category-tree-schema"
        if error_code not in args.ignore_errors:
            log.debug("Validating category_tree.json with JSON schema...")
            category_tree_schema_errors = list(validators.category_tree_validator.iter_errors(category_tree_json))
            if category_tree_schema_errors:
                errors.append({
                    "cause": build_jsonschema_error(category_tree_schema_errors),
                    "error_code": error_code,
                    "message": "Category tree does not conform to schema",
                    "provider_code": provider_code,
                    "location": "category_tree.json",
                })

    return errors


def validate_dataset(dataset_dir):
    errors = []
    provider_code = dataset_dir.storage.provider_code
    dataset_code = dataset_dir.dataset_code

    log.debug("Loading dataset.json...")
    try:
        dataset_json = dataset_dir.load_dataset_json()
    except StorageError as exc:
        errors.append({
            **error_from_storage_error(exc),
            "location": "{}/dataset.json".format(dataset_code),
        })
        return

    # Dataset directory name MUST be the dataset code.
    error_code = "dataset-directory-name"
    if error_code not in args.ignore_errors and dataset_json["code"] != dataset_code:
        errors.append({
            "error_code": error_code,
            "message": "Dataset code from dataset.json ({}) is different than the directory name ({})".format(dataset_json["code"], dataset_code),
            "provider_code": provider_code,
            "dataset_code": dataset_code,
            "location": dataset_code,
        })

    dataset_series = dataset_json.pop("series", None)

    error_code = "dataset-schema"
    if error_code not in args.ignore_errors:
        log.debug("Validating dataset.json with JSON schema (except 'series' property)...")
        dataset_schema_errors = list(validators.dataset_validator.iter_errors(dataset_json))
        if dataset_schema_errors:
            errors.append({
                "cause": build_jsonschema_error(dataset_schema_errors),
                "error_code": error_code,
                "message": "Dataset does not conform to schema",
                "provider_code": provider_code,
                "dataset_code": dataset_code,
                "location": "{}/dataset.json".format(dataset_code),
            })

    return (dataset_json, dataset_series, errors)


def validate_provider(storage):
    """Yield error dicts"""
    errors = []

    log.debug("Loading provider.json...")
    try:
        provider_json = storage.load_provider_json()
    except StorageError as exc:
        errors.append({
            **error_from_storage_error(exc),
            "location": "provider.json",
        })
        return

    provider_code = provider_json.get('code')

    error_code = "provider-schema"
    if error_code not in args.ignore_errors:
        log.debug("Validating provider.json with schema...")
        provider_schema_errors = list(validators.provider_validator.iter_errors(provider_json))
        if provider_schema_errors:
            errors.append({
                "cause": build_jsonschema_error(provider_schema_errors),
                "error_code": error_code,
                "message": "Provider does not conform to schema",
                "provider_code": provider_code,
                "location": "provider.json",
            })

    error_code = "provider-directory-name"
    if error_code not in args.ignore_errors:
        log.debug("Validating provider directory name...")
        error = validate_provider_directory_name(error_code, provider_code)
        if error is not None:
            errors.append(error)

    return (provider_json, errors)


def validate_provider_directory_name(error_code, provider_code=None):
    global args
    if provider_code is None:
        log.debug("Skipped %r validator because provider_code could not be loaded", error_code)
        return None
    provider_code_lower = provider_code.lower()
    valid_directories_names = {
        "{}-json-data".format(provider_code_lower),
        "{}-json-data.git".format(provider_code_lower),
    }
    if args.storage_dir.absolute().name not in valid_directories_names:
        return {
            "error_code": error_code,
            "message": "Directory name must be named one of {}".format(valid_directories_names),
            "provider_code": provider_code,
            "location": ".",
        }
    return None


def validate_observations(provider_code, dataset_code, series_code, location, observations):
    errors = []
    header = observations[0]
    header_nb_columns = len(header)
    rows = observations[1:]

    #
    # Header-level checks
    #

    # The two first columns of the header MUST be named `PERIOD` and `VALUE`.
    error_code = "observations-header-period-value"
    if error_code not in args.ignore_errors:
        if get(0, header, default=None) != "PERIOD":
            errors.append({
                "error_code": error_code,
                "message": "First column of header is not 'PERIOD': {!r}".format(header),
                "provider_code": provider_code,
                "dataset_code": dataset_code,
                "series_code": series_code,
                "location": location,
                "line_number": 0,
            })

        if get(1, header, default=None) != "VALUE":
            errors.append({
                "error_code": error_code,
                "message": "Second column of header is not 'VALUE': {!r}".format(header),
                "provider_code": provider_code,
                "dataset_code": dataset_code,
                "series_code": series_code,
                "location": location,
                "line_number": 0,
            })

    #
    # Column-level checks
    #

    # The `PERIOD` column MUST be sorted in an ascending order.
    error_code = "observations-sorted-periods"
    if error_code not in args.ignore_errors:
        periods = [row[0] for row in rows]
        if periods != sorted(periods):
            errors.append({
                "error_code": error_code,
                "message": "Values of the 'PERIOD' column are not sorted",
                "provider_code": provider_code,
                "dataset_code": dataset_code,
                "series_code": series_code,
                "location": location,
            })

    #
    # Row-level checks
    #

    previous_period_format = None

    for index, observation in enumerate(rows, start=1):
        nb_columns = len(observation)

        # Each row MUST have the same number of columns than the header.
        error_code = "observation-row-size"
        if error_code not in args.ignore_errors and nb_columns != header_nb_columns:
            errors.append({
                "error_code": error_code,
                "message": "Row has {} columns but header has {} columns".format(nb_columns, header_nb_columns),
                "provider_code": provider_code,
                "dataset_code": dataset_code,
                "series_code": series_code,
                "location": location,
                "line_number": index,
            })

        # The values of the `PERIOD` column:
        #   - MUST respect a specific format
        #   - MUST NOT include average values using `M13` or `Q5` periods
        period = observation[0]
        period_format = detect_period_format_strict(period)

        error_code_valid = "observation-period-format-valid"
        error_code_homogeneous = "observations-periods-formats-homogeneous"
        if error_code_valid not in args.ignore_errors and period_format is None:
            errors.append({
                "error_code": error_code_valid,
                "message": "Period format is invalid ({})".format(period),
                "provider_code": provider_code,
                "dataset_code": dataset_code,
                "series_code": series_code,
                "location": location,
                "line_number": index,
            })
        elif error_code_homogeneous not in args.ignore_errors and \
                previous_period_format is not None and previous_period_format != period_format:
            errors.append({
                "error_code": error_code_homogeneous,
                "message": "Period format ({}) is different than previous row ({})".format(period_format, previous_period_format),
                "provider_code": provider_code,
                "dataset_code": dataset_code,
                "series_code": series_code,
                "location": location,
                "line_number": index,
            })

        previous_period_format = period_format

        #  The values of the `VALUE` column MUST either be a float OR be `NA` meaning "not available".
        error_code = "observation-value"
        if error_code not in args.ignore_errors:
            value = observation[1]
            value_float = value_to_float(value)
            if value_float != NOT_AVAILABLE and not isinstance(value_float, float):
                errors.append({
                    "error_code": error_code,
                    "message": "Value is invalid ({})".format(value),
                    "provider_code": provider_code,
                    "dataset_code": dataset_code,
                    "series_code": series_code,
                    "location": location,
                    "line_number": index,
                })

    return errors


def validate_series(dataset_dir, dataset_series):
    errors = []
    provider_code = dataset_dir.storage.provider_code
    dataset_code = dataset_dir.dataset_code

    # Get series from `dataset.json` or `series.jsonl`.
    series_iterator = dataset_dir.iter_series_json(dataset_series=dataset_series)

    if dataset_series and isinstance(dataset_series, list):
        log.debug("Validating series from `dataset.json` with JSON schema...")
        base_path = ["series"]
        base_location = "{}/dataset.json#/series".format(dataset_code)
    else:
        log.debug("Validating series from `series.jsonl` with JSON schema...")
        base_path = []
        base_location = "{}/series.jsonl#".format(dataset_code)

    if args.max_series is not None:
        log.debug("Validating %d series max", args.max_series)
        series_iterator = take(args.max_series, series_iterator)

    validated_series_codes = set()
    validated_series_names = set()
    observations_series_codes = []

    for index, series_json in enumerate(series_iterator):
        series_code = series_json["code"]
        series_name = series_json.get("name")
        series_id_str = dataset_dir.series_id_str(series_code)
        log.debug("Validating series %r...", series_id_str)

        error_code = "series-schema"
        if error_code not in args.ignore_errors:
            series_schema_errors = list(validators.series_validator.iter_errors(series_json))
            if series_schema_errors:
                errors.append({
                    "cause": build_jsonschema_error(series_schema_errors, base_path),
                    "error_code": error_code,
                    "message": "Series does not conform to schema",
                    "provider_code": provider_code,
                    "dataset_code": dataset_code,
                    "series_code": series_code,
                    "location": "{}/{}".format(base_location, index),
                })

        error_code = "duplicate-series-code"
        if error_code not in args.ignore_errors and series_code in validated_series_codes:
            errors.append({
                "error_code": error_code,
                "message": "Series code already met before",
                "provider_code": provider_code,
                "dataset_code": dataset_code,
                "series_code": series_code,
                "location": "{}/{}".format(base_location, index),
            })

        error_code = "duplicate-series-name"
        if error_code not in args.ignore_errors and series_name is not None and series_name in validated_series_names:
            errors.append({
                "error_code": error_code,
                "message": "Series name already met before",
                "provider_code": provider_code,
                "dataset_code": dataset_code,
                "series_code": series_code,
                "location": "{}/{}".format(base_location, index),
            })

        observations = series_json.get("observations")
        if observations is None:
            # Validate observations in TSV files later.
            observations_series_codes.append(series_code)
        else:
            # Validate observations in series JSON lines file now, to avoid loading them again later.
            series_jsonl_file_name = dataset_dir.get_series_jsonl_file_name(dataset_series=dataset_series)
            if series_jsonl_file_name is None:
                error_code = "observations-in-dataset-json"
                if error_code not in args.ignore_errors:
                    errors.append({
                        "error_code": error_code,
                        "message": "Having observations in 'series' property of dataset.json is not supported.",
                        "provider_code": provider_code,
                        "dataset_code": dataset_code,
                        "series_code": series_code,
                        "location": "{}/dataset.json".format(dataset_code),
                    })
            else:
                series_file_path = dataset_dir.path / series_jsonl_file_name
                log.debug("Validating observations (in JSON lines format) of series %r...", series_id_str)
                location = str(series_file_path.relative_to(dataset_dir.storage.path))
                observations_errors = validate_observations(
                    provider_code, dataset_code, series_code, location, observations)
                errors.extend(observations_errors)

        validated_series_codes.add(series_code)
        validated_series_names.add(series_name)

    if observations_series_codes:
        log.debug("Validating observations (in TSV format) of %d series...", len(observations_series_codes))

        observations_iterator = dataset_dir.iter_observations(observations_series_codes)

        if args.max_observations is not None:
            log.debug("Validating %d observations max", args.max_observations)
            observations_iterator = take(args.max_observations, observations_iterator)

        for series_code, file_path, observations in observations_iterator:
            series_id_str = dataset_dir.series_id_str(series_code)
            log.debug("Validating observations of series %r...", series_id_str)
            location = str(file_path.relative_to(dataset_dir.storage.path))
            observations_errors = validate_observations(
                provider_code, dataset_code, series_code, location, observations)
            errors.extend(observations_errors)

    return (series_json, errors)


if __name__ == '__main__':
    sys.exit(main())
